{-# OPTIONS -#include "Curses_hsc.h" #-}
{-# LINE 1 "Curses.hsc" #-}
module Curses (
{-# LINE 2 "Curses.hsc" #-}
    --------------------------------------------------------------------
    
    Window,      -- data Window deriving Eq
    stdScr,      -- :: Window
    initScr,     -- :: IO Window
    cBreak,      -- :: Bool -> IO ()
    raw,         -- :: Bool -> IO ()
    echo,        -- :: Bool -> IO ()
    nl,          -- :: Bool -> IO ()
    intrFlush,   -- :: Bool -> IO ()
    keypad,      -- :: Window -> Bool -> IO ()
    initCurses,  -- :: IO ()
    endWin,      -- :: IO ()
    
    --------------------------------------------------------------------
    
    scrSize,          -- :: IO (Int, Int)
    refresh,          -- :: IO ()
    
    --------------------------------------------------------------------
    
    hasColors,      -- :: IO Bool
    startColor,     -- :: IO ()
    Pair(..),       -- newtype Pair = Pair Int deriving (Eq, Ord, Ix)
    colorPairs,     -- :: IO Int
    Color(..),      -- newtype Color = Color Int deriving (Eq, Ord, Ix)
    colors,         -- :: IO Int
    black, red, green, yellow, blue, magenta, cyan, white, -- :: Color
    initPair,       -- :: Pair -> Color -> Color -> IO ()
    pairContent,    -- :: Pair -> IO (Color, Color)
    canChangeColor, -- :: IO Bool
    initColor,      -- :: Color -> (Int, Int, Int) -> IO ()
    colorContent,   -- :: Color -> IO (Int, Int, Int)
    
    --------------------------------------------------------------------
    
    Attr,  -- data Attr deriving Eq
    attr0, -- :: Attr
    
    isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis,
    isLeft, isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical,
        -- :: Attr -> Bool
    
    setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical,
        -- :: Attr -> Bool -> Attr
    
    attrSet, -- :: Attr -> Pair -> IO ()
    attrOn, attrOff,
    
    --------------------------------------------------------------------

    ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee,
    hLine, vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus,
    bullet, lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    -- The rest of characters is not standard. ncurses says:
    -- "These aren't documented, but a lot of System Vs have them
    -- anyway (you can spot pprryyzz{{||}} in a lot of AT&T terminfo
    -- strings). The ACS_names may not match AT&T's, our source
    -- didn't know them."
    s3, s7, lEqual, gEqual, pi, nEqual, sterling,
        -- :: Char
    wAddStr, 
    addStr,        -- :: String -> IO ()
    addStrLn,      -- :: String -> IO ()
    addGraphStr,   -- :: String -> IO ()
    addGraphStrLn, -- :: String -> IO ()
    addLn,         -- :: IO ()
    
    --------------------------------------------------------------------
    
    bkgrndSet,      -- :: Attr -> Pair -> IO ()
    erase,          -- :: IO ()
    clrToEol,       -- :: IO ()
    move,           -- :: Int -> Int -> IO ()
    Visibility(..), -- data Visibility = Invisible | Visible | VeryVisible
    cursSet,        -- :: Visibility -> IO Visibility
    standout,standend,
    attrDim, attrBold,
    attrDimOn, attrDimOff,
    attrBoldOn, attrBoldOff,
    wAttrOn,
    wAttrOff, 
    touchWin,
    mvWAddStr,
    wMove,
    
    --------------------------------------------------------------------
    
    Key(..),
    getCh, getChHack,
    newPad, pRefresh, delWin, newWin,
    wClrToEol,
    withProgram,
    beep, wAttrSet, wAttrGet
    ) 
    
    --------------------------------------------------------------------
    where

import Prelude hiding (pi)
import Monad 
import Char           (chr, ord)
import Ix             (Ix)
import Bits
import Concurrent

import Foreign
import CForeign

import GHC.IOBase
import Control.Exception hiding(block)

import Charset



{-# LINE 121 "Curses.hsc" #-}


------------------------------------------------------------------------

fi = fromIntegral

throwIfErr :: Num a => String -> IO a -> IO a
throwIfErr name act = do
    res <- act
    if res == (-1)
{-# LINE 131 "Curses.hsc" #-}
        then ioError (userError ("Curses: "++name++" failed"))
        else return res

throwIfErr_ :: Num a => String -> IO a -> IO ()
throwIfErr_ name act = void $ throwIfErr name act

------------------------------------------------------------------------

data WindowTag 
type Window = Ptr WindowTag

stdScr :: Window
stdScr = unsafePerformIO (peek stdscr)
foreign import ccall "static my_curses.h &stdscr" stdscr :: Ptr Window


initScr :: IO Window
initScr = throwIfNull "initscr" initscr
foreign import ccall unsafe "my_curses.h initscr" initscr :: IO Window

cBreak :: Bool -> IO ()
cBreak True  = throwIfErr_ "cbreak"   cbreak
cBreak False = throwIfErr_ "nocbreak" nocbreak
foreign import ccall unsafe "my_curses.h cbreak" cbreak :: IO CInt
foreign import ccall unsafe "my_curses.h nocbreak" nocbreak :: IO CInt

raw :: Bool -> IO ()
raw False = throwIfErr_ "noraw" noraw
raw True  = throwIfErr_ "raw"   raw_c
foreign import ccall unsafe "my_curses.h noraw" noraw :: IO CInt
foreign import ccall unsafe "my_curses.h raw" raw_c :: IO CInt

echo :: Bool -> IO ()
echo False = throwIfErr_ "noecho" noecho
echo True  = throwIfErr_ "echo"   echo_c
foreign import ccall unsafe "my_curses.h noecho" noecho :: IO CInt
foreign import ccall unsafe "my_curses.h echo" echo_c :: IO CInt

nl :: Bool -> IO ()
nl True  = throwIfErr_ "nl"   nl_c
nl False = throwIfErr_ "nonl" nonl
foreign import ccall unsafe "my_curses.h nl" nl_c :: IO CInt
foreign import ccall unsafe "my_curses.h nonl" nonl :: IO CInt

intrFlush :: Bool -> IO ()
intrFlush bf =
    throwIfErr_ "intrflush" $ intrflush stdScr (if bf then 1 else 0)
foreign import ccall unsafe intrflush :: Window -> (Word32) -> IO CInt
{-# LINE 179 "Curses.hsc" #-}

keypad :: Window -> Bool -> IO ()
keypad win bf =
    throwIfErr_ "keypad" $ keypad_c win (if bf then 1 else 0)
foreign import ccall unsafe "keypad" keypad_c :: Window -> (Word32) -> IO CInt
{-# LINE 184 "Curses.hsc" #-}

initCurses :: IO ()
initCurses = do
    initScr; cBreak True; echo False
    nl False; intrFlush False; keypad stdScr True

endWin :: IO ()
endWin = throwIfErr_ "endwin" endwin
foreign import ccall unsafe endwin :: IO CInt

------------------------------------------------------------------------

scrSize :: IO (Int, Int)
scrSize = do
    lines <- peek linesPtr
    cols  <- peek colsPtr
    return (fromIntegral lines, fromIntegral cols)

foreign import ccall "my_curses.h &LINES" linesPtr :: Ptr CInt
foreign import ccall "my_curses.h &COLS" colsPtr :: Ptr CInt


refresh :: IO ()
refresh = throwIfErr_ "refresh" refresh_c
foreign import ccall unsafe "refresh" refresh_c :: IO CInt

------------------------------------------------------------------------

hasColors :: IO Bool
hasColors = liftM (/= 0) has_colors
foreign import ccall unsafe has_colors :: IO (Word32)
{-# LINE 215 "Curses.hsc" #-}

startColor :: IO ()
startColor = throwIfErr_ "start_color" start_color
foreign import ccall unsafe start_color :: IO CInt

newtype Pair = Pair Int deriving (Eq, Ord, Ix)

colorPairs :: IO Int
colorPairs = fmap fromIntegral $ peek colorPairsPtr


foreign import ccall "my_curses.h &COLOR_PAIRS" colorPairsPtr :: Ptr CInt

newtype Color = Color Int deriving (Eq, Ord, Ix)

colors :: IO Int
colors = liftM fromIntegral $ peek colorsPtr

foreign import ccall "my_curses.h &COLORS" colorsPtr :: Ptr CInt

black, red, green, yellow, blue, magenta, cyan, white :: Color
black   = Color (0)
{-# LINE 237 "Curses.hsc" #-}
red     = Color (1)
{-# LINE 238 "Curses.hsc" #-}
green   = Color (2)
{-# LINE 239 "Curses.hsc" #-}
yellow  = Color (3)
{-# LINE 240 "Curses.hsc" #-}
blue    = Color (4)
{-# LINE 241 "Curses.hsc" #-}
magenta = Color (5)
{-# LINE 242 "Curses.hsc" #-}
cyan    = Color (6)
{-# LINE 243 "Curses.hsc" #-}
white   = Color (7)
{-# LINE 244 "Curses.hsc" #-}

initPair :: Pair -> Color -> Color -> IO ()
initPair (Pair p) (Color f) (Color b) =
    throwIfErr_ "init_pair" $
        init_pair (fromIntegral p) (fromIntegral f) (fromIntegral b)
foreign import ccall unsafe init_pair :: CShort -> CShort -> CShort -> IO CInt

pairContent :: Pair -> IO (Color, Color)
pairContent (Pair p) =
    alloca $ \fPtr ->
    alloca $ \bPtr -> do
        throwIfErr "pair_content" $ pair_content (fromIntegral p) fPtr bPtr
        f <- peek fPtr
        b <- peek bPtr
        return (Color (fromIntegral f), Color (fromIntegral b))
foreign import ccall unsafe pair_content :: CShort -> Ptr CShort -> Ptr CShort -> IO CInt

canChangeColor :: IO Bool
canChangeColor = liftM (/= 0) can_change_color
foreign import ccall unsafe can_change_color :: IO (Word32)
{-# LINE 264 "Curses.hsc" #-}

initColor :: Color -> (Int, Int, Int) -> IO ()
initColor (Color c) (r, g, b) =
    throwIfErr_ "init_color" $
        init_color (fromIntegral c) (fromIntegral r) (fromIntegral g) (fromIntegral b)
foreign import ccall unsafe init_color :: CShort -> CShort -> CShort -> CShort -> IO CInt

colorContent :: Color -> IO (Int, Int, Int)
colorContent (Color c) =
    alloca $ \rPtr ->
    alloca $ \gPtr ->
    alloca $ \bPtr -> do
        throwIfErr "color_content" $ color_content (fromIntegral c) rPtr gPtr bPtr
        r <- peek rPtr
        g <- peek gPtr
        b <- peek bPtr
        return (fromIntegral r, fromIntegral g, fromIntegral b)
foreign import ccall unsafe color_content :: CShort -> Ptr CShort -> Ptr CShort -> Ptr CShort -> IO CInt

foreign import ccall unsafe "my_curses.h hs_curses_color_pair" colorPair :: Pair -> (Word32)
{-# LINE 284 "Curses.hsc" #-}

{-# LINE 285 "Curses.hsc" #-}

-------------
-- Attributes 
-------------

foreign import ccall unsafe "my_curses.h attr_set" attr_set :: Attr -> CShort -> Ptr a -> IO Int
-- foreign import ccall unsafe "my_curses.h attr_get" :: Attr -> CShort -> Ptr a -> IO Int

foreign import ccall unsafe "my_curses.h wattr_set" wattr_set :: Window -> Attr -> CInt -> Ptr a -> IO CInt
foreign import ccall unsafe "my_curses.h wattr_get" wattr_get :: Window -> Ptr Attr -> Ptr CShort -> Ptr a -> IO CInt

foreign import ccall "my_curses.h attr_on" attr_on :: (Word32) -> Ptr a -> IO Int
{-# LINE 297 "Curses.hsc" #-}
foreign import ccall "my_curses.h attr_off" attr_off :: (Word32) -> Ptr a -> IO Int
{-# LINE 298 "Curses.hsc" #-}
foreign import ccall "my_curses.h attron" attron :: Int -> IO Int
foreign import ccall "my_curses.h attroff" attroff :: Int -> IO Int
foreign import ccall unsafe "my_curses.h wattron" wattron :: Window -> CInt -> IO CInt
foreign import ccall unsafe "my_curses.h wattroff" wattroff :: Window -> CInt -> IO CInt
foreign import ccall standout :: IO Int
foreign import ccall standend :: IO Int

wAttrSet :: Window -> (Attr,Pair) -> IO ()
wAttrSet w (a,(Pair p)) = throwIfErr_ "wattr_set" $ wattr_set w a (fromIntegral p) nullPtr
wAttrGet :: Window -> IO (Attr,Pair)
wAttrGet w =  alloca $ \pa -> alloca $ \pp -> (throwIfErr_ "wattr_get" $ wattr_get w pa pp nullPtr) >> (peek pa >>=  \a -> peek pp >>=  \p -> return (a,Pair (fromIntegral p)))


newtype Attr = Attr (Word32) deriving (Eq,Storable,Bits)
{-# LINE 312 "Curses.hsc" #-}

attr0 :: Attr
attr0 = Attr (0)
{-# LINE 315 "Curses.hsc" #-}

isAltCharset, isBlink, isBold, isDim, isHorizontal, isInvis, isLeft,
    isLow, isProtect, isReverse, isRight, isStandout, isTop,
    isUnderline, isVertical
    :: Attr -> Bool
isAltCharset = isAttr (4194304)
{-# LINE 321 "Curses.hsc" #-}
isBlink      = isAttr (524288)
{-# LINE 322 "Curses.hsc" #-}
isBold       = isAttr (2097152)
{-# LINE 323 "Curses.hsc" #-}
isDim        = isAttr (1048576)
{-# LINE 324 "Curses.hsc" #-}
isHorizontal = isAttr (33554432)
{-# LINE 325 "Curses.hsc" #-}
isInvis      = isAttr (8388608)
{-# LINE 326 "Curses.hsc" #-}
isLeft       = isAttr (67108864)
{-# LINE 327 "Curses.hsc" #-}
isLow        = isAttr (134217728)
{-# LINE 328 "Curses.hsc" #-}
isProtect    = isAttr (16777216)
{-# LINE 329 "Curses.hsc" #-}
isReverse    = isAttr (262144)
{-# LINE 330 "Curses.hsc" #-}
isRight      = isAttr (268435456)
{-# LINE 331 "Curses.hsc" #-}
isStandout   = isAttr (65536)
{-# LINE 332 "Curses.hsc" #-}
isTop        = isAttr (536870912)
{-# LINE 333 "Curses.hsc" #-}
isUnderline  = isAttr (131072)
{-# LINE 334 "Curses.hsc" #-}
isVertical   = isAttr (1073741824)
{-# LINE 335 "Curses.hsc" #-}

isAttr :: (Word32) -> Attr -> Bool
{-# LINE 337 "Curses.hsc" #-}
isAttr bit (Attr a) = a .&. bit /= 0

setAltCharset, setBlink, setBold, setDim, setHorizontal, setInvis,
    setLeft, setLow, setProtect, setReverse, setRight, setStandout,
    setTop, setUnderline, setVertical
    :: Attr -> Bool -> Attr
setAltCharset = setAttr (4194304)
{-# LINE 344 "Curses.hsc" #-}
setBlink      = setAttr (524288)
{-# LINE 345 "Curses.hsc" #-}
setBold       = setAttr (2097152)
{-# LINE 346 "Curses.hsc" #-}
setDim        = setAttr (1048576)
{-# LINE 347 "Curses.hsc" #-}
setHorizontal = setAttr (33554432)
{-# LINE 348 "Curses.hsc" #-}
setInvis      = setAttr (8388608)
{-# LINE 349 "Curses.hsc" #-}
setLeft       = setAttr (67108864)
{-# LINE 350 "Curses.hsc" #-}
setLow        = setAttr (134217728)
{-# LINE 351 "Curses.hsc" #-}
setProtect    = setAttr (16777216)
{-# LINE 352 "Curses.hsc" #-}
setReverse    = setAttr (262144)
{-# LINE 353 "Curses.hsc" #-}
setRight      = setAttr (268435456)
{-# LINE 354 "Curses.hsc" #-}
setStandout   = setAttr (65536)
{-# LINE 355 "Curses.hsc" #-}
setTop        = setAttr (536870912)
{-# LINE 356 "Curses.hsc" #-}
setUnderline  = setAttr (131072)
{-# LINE 357 "Curses.hsc" #-}
setVertical   = setAttr (1073741824)
{-# LINE 358 "Curses.hsc" #-}

setAttr :: (Word32) -> Attr -> Bool -> Attr
{-# LINE 360 "Curses.hsc" #-}
setAttr bit (Attr a) False = Attr (a .&. complement bit)
setAttr bit (Attr a) True  = Attr (a .|.            bit)

attrSet :: Attr -> Pair -> IO ()
attrSet attr (Pair p) = throwIfErr_ "attrset" $
    attr_set attr (fromIntegral p) nullPtr

attrOn :: Attr -> IO ()
attrOn (Attr attr) = throwIfErr_ "attr_on" $
    attr_on attr nullPtr


attrOff :: Attr -> IO ()
attrOff (Attr attr) = throwIfErr_ "attr_off" $
    attr_off attr nullPtr



wAttrOn :: Window -> Int -> IO ()
wAttrOn w x = throwIfErr_ "wattron" $ wattron w (fi x)

wAttrOff :: Window -> Int -> IO ()
wAttrOff w x = throwIfErr_ "wattroff" $ wattroff w (fi x)

attrDimOn :: IO ()
attrDimOn  = throwIfErr_ "attron A_DIM" $
    attron (1048576) 
{-# LINE 387 "Curses.hsc" #-}

attrDimOff :: IO ()
attrDimOff = throwIfErr_ "attroff A_DIM" $
    attroff (1048576) 
{-# LINE 391 "Curses.hsc" #-}

attrBoldOn :: IO ()
attrBoldOn  = throwIfErr_ "attron A_BOLD" $
    attron (2097152) 
{-# LINE 395 "Curses.hsc" #-}

attrBoldOff :: IO ()
attrBoldOff = throwIfErr_ "attroff A_BOLD" $
    attroff (2097152) 
{-# LINE 399 "Curses.hsc" #-}


attrDim :: Int
attrDim = (1048576)
{-# LINE 403 "Curses.hsc" #-}
attrBold :: Int
attrBold = (2097152)
{-# LINE 405 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 596 "Curses.hsc" #-}

ulCorner, llCorner, urCorner, lrCorner, rTee, lTee, bTee, tTee, hLine,
    vLine, plus, s1, s9, diamond, ckBoard, degree, plMinus, bullet,
    lArrow, rArrow, dArrow, uArrow, board, lantern, block,
    s3, s7, lEqual, gEqual, pi, nEqual, sterling
    :: Char

ulCorner = '+'
llCorner = '+'
urCorner = '+'
lrCorner = '+'
rTee     = '+'
lTee     = '+'
bTee     = '+'
tTee     = '+'
hLine    = '-'
vLine    = '|'
plus     = '+'
s1       = '-'
s9       = '_'
diamond  = '+'
ckBoard  = ':'
degree   = '\''
plMinus  = '#'
bullet   = 'o'
lArrow   = '<'
rArrow   = '>'
dArrow   = 'v'
uArrow   = '^'
board    = '#'
lantern  = '#'
block    = '#'
s3       = '-'
s7       = '-'
lEqual   = '<'
gEqual   = '>'
pi       = '*'
nEqual   = '!'
sterling = 'f'


-- setCursesOutConv :: ([Char] -> [Word8]) -> IO ()
-- setCursesOutConv newConv = writeIORef cursesOutConv (withCConvString newConv)

-- {-# NOINLINE cursesOutConv #-}
cursesOutConv :: IORef (String -> (Ptr CChar -> IO b) -> IO b)
cursesOutConv = unsafePerformIO $ newIORef conv where
    conv s f = withArray0 0 (stringToBytes s) (f . castPtr)

-- withCUTFString :: String -> (Ptr a -> IO b) -> IO b
-- withCUTFString s f = withArray0 0 (toUTF s) (f . castPtr)

withCConvString :: ([Char] -> [Word8]) -> String -> (Ptr CChar -> IO b) -> IO b
withCConvString conv s f = withArray0 0 (conv s) (f . castPtr)


addStr :: String -> IO ()
addStr str = do
	conv <- readIORef cursesOutConv
	throwIfErr_ "addstr" $ conv str addstr
foreign import ccall unsafe addstr :: Ptr CChar -> IO CInt

wAddStr :: Window -> String -> IO ()
wAddStr w str = do
	conv <- readIORef cursesOutConv
	throwIfErr_ "waddstr" $ conv str (waddstr w)

mvWAddStr :: Window -> Int -> Int -> String -> IO ()
mvWAddStr w y x str = do
	conv <- readIORef cursesOutConv
	conv str (mvwaddstr w (fi y) (fi x))
	return ()
	--throwIfErr_ "mvwaddstr" $ conv str (mvwaddstr w (fi y) (fi x))

foreign import ccall unsafe waddstr :: Window -> Ptr CChar -> IO CInt
foreign import ccall unsafe mvwaddstr :: Window -> CInt -> CInt -> Ptr CChar -> IO CInt

addStrLn :: String -> IO ()
addStrLn str = do addStr str; addLn

addGraphStr :: String -> IO ()
addGraphStr = addStr

addGraphStrLn :: String -> IO ()
addGraphStrLn = addStrLn


{-# LINE 683 "Curses.hsc" #-}

addLn :: IO ()
addLn = throwIfErr_ "addch" $ addch (fromIntegral (ord '\n'))

foreign import ccall unsafe addch :: (Word32) -> IO CInt
{-# LINE 688 "Curses.hsc" #-}

------------------------------------------------------------------------


{-# LINE 694 "Curses.hsc" #-}

bkgrndSet :: Attr -> Pair -> IO ()
bkgrndSet (Attr a) p = bkgdset $
    fromIntegral (ord ' ') .|.
    (if a .&. 4194304 /= 0 then 4194304 else 0) .|.
{-# LINE 699 "Curses.hsc" #-}
    (if a .&. 524288 /= 0 then 524288 else 0) .|.
{-# LINE 700 "Curses.hsc" #-}
    (if a .&. 2097152 /= 0 then 2097152 else 0) .|.
{-# LINE 701 "Curses.hsc" #-}
    (if a .&. 1048576 /= 0 then 1048576 else 0) .|.
{-# LINE 702 "Curses.hsc" #-}
    (if a .&. 8388608 /= 0 then 8388608 else 0) .|.
{-# LINE 703 "Curses.hsc" #-}
    (if a .&. 16777216 /= 0 then 16777216 else 0) .|.
{-# LINE 704 "Curses.hsc" #-}
    (if a .&. 262144 /= 0 then 262144 else 0) .|.
{-# LINE 705 "Curses.hsc" #-}
    (if a .&. 65536 /= 0 then 65536 else 0) .|.
{-# LINE 706 "Curses.hsc" #-}
    (if a .&. 131072 /= 0 then 131072 else 0) .|.
{-# LINE 707 "Curses.hsc" #-}
    colorPair p
foreign import ccall unsafe bkgdset :: (Word32) -> IO ()
{-# LINE 709 "Curses.hsc" #-}

erase :: IO ()
erase = throwIfErr_ "erase" erase_c
foreign import ccall unsafe "erase" erase_c :: IO CInt

clrToEol :: IO ()
clrToEol = throwIfErr_ "clrtoeol" clrtoeol
foreign import ccall unsafe clrtoeol :: IO CInt

move :: Int -> Int -> IO ()
move y x =
    throwIfErr_ "move" $ move_c (fromIntegral y) (fromIntegral x)
foreign import ccall unsafe "move" move_c :: CInt -> CInt -> IO CInt

wMove :: Window -> Int -> Int -> IO ()
wMove w y x = throwIfErr_ "wmove" $ wmove w (fi y) (fi x)
foreign import ccall unsafe  wmove :: Window -> CInt -> CInt -> IO CInt

data Visibility = Invisible | Visible | VeryVisible

cursSet :: Visibility -> IO Visibility
cursSet vis = do
    oldVis_c <- throwIfErr "curs_set" $ curs_set vis_c
    case  oldVis_c of
        0 -> return Invisible
        1 -> return Visible
        2 -> return VeryVisible
        _ -> ioError $ userError $
            "Curses: Unexpected curs_set result ("++show oldVis_c++")"
    where
    vis_c :: CInt
    vis_c = case vis of
        Invisible   -> 0
        Visible     -> 1
        VeryVisible -> 2
foreign import ccall unsafe curs_set :: CInt -> IO CInt


------------------------------------------------------------------------


touchWin :: Window -> IO ()
touchWin w = throwIfErr_ "touchwin" $ touchwin w
foreign import ccall touchwin :: Window -> IO CInt

newPad :: Int -> Int -> IO Window
newPad nlines ncols = throwIfNull "newpad" $ newpad (fromIntegral nlines) (fromIntegral ncols)

pRefresh :: Window -> Int -> Int -> Int -> Int -> Int -> Int -> IO ()
pRefresh pad pminrow pmincol sminrow smincol smaxrow smaxcol = throwIfErr_ "prefresh" $
    prefresh pad (fromIntegral pminrow) (fromIntegral pmincol) (fromIntegral sminrow) (fromIntegral smincol) (fromIntegral smaxrow) (fromIntegral smaxcol)

delWin :: Window -> IO ()
delWin w = throwIfErr_ "delwin" $ delwin w
    
foreign import ccall unsafe prefresh :: Window -> CInt -> CInt -> CInt -> CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe newpad :: CInt -> CInt -> IO Window
foreign import ccall unsafe delwin :: Window -> IO CInt


newWin :: Int -> Int -> Int -> Int -> IO Window
newWin nlines ncolumn begin_y begin_x = throwIfNull "newwin" $ newwin (fi nlines) (fi ncolumn) (fi begin_y) (fi begin_x)

foreign import ccall unsafe newwin :: CInt -> CInt -> CInt -> CInt -> IO Window


wClrToEol :: Window -> IO ()
wClrToEol w = throwIfErr_ "wclrtoeol" $ wclrtoeol w
foreign import ccall unsafe wclrtoeol :: Window -> IO CInt




-- foreign import ccall unsafe halfdelay :: CInt -> IO CInt
foreign import ccall threadsafe getch :: IO CInt
-- foreign import ccall unsafe nodelay :: Window -> (#type bool) -> IO CInt


foreign import ccall unsafe def_prog_mode :: IO CInt
foreign import ccall unsafe reset_prog_mode :: IO CInt


withProgram :: IO a -> IO a
withProgram action = Control.Exception.bracket_ (def_prog_mode >> endWin) reset_prog_mode action


foreign import ccall unsafe "my_curses.h beep" c_beep :: IO CInt
foreign import ccall unsafe "my_curses.h flash" c_flash :: IO CInt

beep :: IO ()
beep = do
    br <- c_beep
    when (br /= (0)) (c_flash >> return ()) 
{-# LINE 802 "Curses.hsc" #-}


---------------
-- Key Routines
---------------

data Key
    = KeyChar Char | KeyBreak | KeyDown | KeyUp | KeyLeft | KeyRight
    | KeyHome | KeyBackspace | KeyF Int | KeyDL | KeyIL | KeyDC
    | KeyIC | KeyEIC | KeyClear | KeyEOS | KeyEOL | KeySF | KeySR
    | KeyNPage | KeyPPage | KeySTab | KeyCTab | KeyCATab | KeyEnter
    | KeySReset | KeyReset | KeyPrint | KeyLL | KeyA1 | KeyA3
    | KeyB2 | KeyC1 | KeyC3 | KeyBTab | KeyBeg | KeyCancel | KeyClose
    | KeyCommand | KeyCopy | KeyCreate | KeyEnd | KeyExit | KeyFind
    | KeyHelp | KeyMark | KeyMessage | KeyMove | KeyNext | KeyOpen
    | KeyOptions | KeyPrevious | KeyRedo | KeyReference | KeyRefresh
    | KeyReplace | KeyRestart | KeyResume | KeySave | KeySBeg
    | KeySCancel | KeySCommand | KeySCopy | KeySCreate | KeySDC
    | KeySDL | KeySelect | KeySEnd | KeySEOL | KeySExit | KeySFind
    | KeySHelp | KeySHome | KeySIC | KeySLeft | KeySMessage | KeySMove
    | KeySNext | KeySOptions | KeySPrevious | KeySPrint | KeySRedo
    | KeySReplace | KeySRight | KeySRsume | KeySSave | KeySSuspend
    | KeySUndo | KeySuspend | KeyUndo | KeyResize | KeyUnknown Int
    deriving (Eq,Show)

decodeKey :: CInt -> Key
decodeKey key = case key of
    _ | key >= 0 && key <= 255 -> KeyChar (chr (fromIntegral key))
    (257)         -> KeyBreak
{-# LINE 831 "Curses.hsc" #-}
    (258)          -> KeyDown
{-# LINE 832 "Curses.hsc" #-}
    (259)            -> KeyUp
{-# LINE 833 "Curses.hsc" #-}
    (260)          -> KeyLeft
{-# LINE 834 "Curses.hsc" #-}
    (261)         -> KeyRight
{-# LINE 835 "Curses.hsc" #-}
    (262)          -> KeyHome
{-# LINE 836 "Curses.hsc" #-}
    (263)     -> KeyBackspace
{-# LINE 837 "Curses.hsc" #-}
    _ | key >= (264) && key <= (327)
{-# LINE 838 "Curses.hsc" #-}
                               -> KeyF (fromIntegral (key - 264))
{-# LINE 839 "Curses.hsc" #-}
    (328)            -> KeyDL
{-# LINE 840 "Curses.hsc" #-}
    (329)            -> KeyIL
{-# LINE 841 "Curses.hsc" #-}
    (330)            -> KeyDC
{-# LINE 842 "Curses.hsc" #-}
    (331)            -> KeyIC
{-# LINE 843 "Curses.hsc" #-}
    (332)           -> KeyEIC
{-# LINE 844 "Curses.hsc" #-}
    (333)         -> KeyClear
{-# LINE 845 "Curses.hsc" #-}
    (334)           -> KeyEOS
{-# LINE 846 "Curses.hsc" #-}
    (335)           -> KeyEOL
{-# LINE 847 "Curses.hsc" #-}
    (336)            -> KeySF
{-# LINE 848 "Curses.hsc" #-}
    (337)            -> KeySR
{-# LINE 849 "Curses.hsc" #-}
    (338)         -> KeyNPage
{-# LINE 850 "Curses.hsc" #-}
    (339)         -> KeyPPage
{-# LINE 851 "Curses.hsc" #-}
    (340)          -> KeySTab
{-# LINE 852 "Curses.hsc" #-}
    (341)          -> KeyCTab
{-# LINE 853 "Curses.hsc" #-}
    (342)         -> KeyCATab
{-# LINE 854 "Curses.hsc" #-}
    (343)         -> KeyEnter
{-# LINE 855 "Curses.hsc" #-}
    (344)        -> KeySReset
{-# LINE 856 "Curses.hsc" #-}
    (345)         -> KeyReset
{-# LINE 857 "Curses.hsc" #-}
    (346)         -> KeyPrint
{-# LINE 858 "Curses.hsc" #-}
    (347)            -> KeyLL
{-# LINE 859 "Curses.hsc" #-}
    (348)            -> KeyA1
{-# LINE 860 "Curses.hsc" #-}
    (349)            -> KeyA3
{-# LINE 861 "Curses.hsc" #-}
    (350)            -> KeyB2
{-# LINE 862 "Curses.hsc" #-}
    (351)            -> KeyC1
{-# LINE 863 "Curses.hsc" #-}
    (352)            -> KeyC3
{-# LINE 864 "Curses.hsc" #-}
    (353)          -> KeyBTab
{-# LINE 865 "Curses.hsc" #-}
    (354)           -> KeyBeg
{-# LINE 866 "Curses.hsc" #-}
    (355)        -> KeyCancel
{-# LINE 867 "Curses.hsc" #-}
    (356)         -> KeyClose
{-# LINE 868 "Curses.hsc" #-}
    (357)       -> KeyCommand
{-# LINE 869 "Curses.hsc" #-}
    (358)          -> KeyCopy
{-# LINE 870 "Curses.hsc" #-}
    (359)        -> KeyCreate
{-# LINE 871 "Curses.hsc" #-}
    (360)           -> KeyEnd
{-# LINE 872 "Curses.hsc" #-}
    (361)          -> KeyExit
{-# LINE 873 "Curses.hsc" #-}
    (362)          -> KeyFind
{-# LINE 874 "Curses.hsc" #-}
    (363)          -> KeyHelp
{-# LINE 875 "Curses.hsc" #-}
    (364)          -> KeyMark
{-# LINE 876 "Curses.hsc" #-}
    (365)       -> KeyMessage
{-# LINE 877 "Curses.hsc" #-}
    (366)          -> KeyMove
{-# LINE 878 "Curses.hsc" #-}
    (367)          -> KeyNext
{-# LINE 879 "Curses.hsc" #-}
    (368)          -> KeyOpen
{-# LINE 880 "Curses.hsc" #-}
    (369)       -> KeyOptions
{-# LINE 881 "Curses.hsc" #-}
    (370)      -> KeyPrevious
{-# LINE 882 "Curses.hsc" #-}
    (371)          -> KeyRedo
{-# LINE 883 "Curses.hsc" #-}
    (372)     -> KeyReference
{-# LINE 884 "Curses.hsc" #-}
    (373)       -> KeyRefresh
{-# LINE 885 "Curses.hsc" #-}
    (374)       -> KeyReplace
{-# LINE 886 "Curses.hsc" #-}
    (375)       -> KeyRestart
{-# LINE 887 "Curses.hsc" #-}
    (376)        -> KeyResume
{-# LINE 888 "Curses.hsc" #-}
    (377)          -> KeySave
{-# LINE 889 "Curses.hsc" #-}
    (378)          -> KeySBeg
{-# LINE 890 "Curses.hsc" #-}
    (379)       -> KeySCancel
{-# LINE 891 "Curses.hsc" #-}
    (380)      -> KeySCommand
{-# LINE 892 "Curses.hsc" #-}
    (381)         -> KeySCopy
{-# LINE 893 "Curses.hsc" #-}
    (382)       -> KeySCreate
{-# LINE 894 "Curses.hsc" #-}
    (383)           -> KeySDC
{-# LINE 895 "Curses.hsc" #-}
    (384)           -> KeySDL
{-# LINE 896 "Curses.hsc" #-}
    (385)        -> KeySelect
{-# LINE 897 "Curses.hsc" #-}
    (386)          -> KeySEnd
{-# LINE 898 "Curses.hsc" #-}
    (387)          -> KeySEOL
{-# LINE 899 "Curses.hsc" #-}
    (388)         -> KeySExit
{-# LINE 900 "Curses.hsc" #-}
    (389)         -> KeySFind
{-# LINE 901 "Curses.hsc" #-}
    (390)         -> KeySHelp
{-# LINE 902 "Curses.hsc" #-}
    (391)         -> KeySHome
{-# LINE 903 "Curses.hsc" #-}
    (392)           -> KeySIC
{-# LINE 904 "Curses.hsc" #-}
    (393)         -> KeySLeft
{-# LINE 905 "Curses.hsc" #-}
    (394)      -> KeySMessage
{-# LINE 906 "Curses.hsc" #-}
    (395)         -> KeySMove
{-# LINE 907 "Curses.hsc" #-}
    (396)         -> KeySNext
{-# LINE 908 "Curses.hsc" #-}
    (397)      -> KeySOptions
{-# LINE 909 "Curses.hsc" #-}
    (398)     -> KeySPrevious
{-# LINE 910 "Curses.hsc" #-}
    (399)        -> KeySPrint
{-# LINE 911 "Curses.hsc" #-}
    (400)         -> KeySRedo
{-# LINE 912 "Curses.hsc" #-}
    (401)      -> KeySReplace
{-# LINE 913 "Curses.hsc" #-}
    (402)        -> KeySRight
{-# LINE 914 "Curses.hsc" #-}
    (403)        -> KeySRsume
{-# LINE 915 "Curses.hsc" #-}
    (404)         -> KeySSave
{-# LINE 916 "Curses.hsc" #-}
    (405)      -> KeySSuspend
{-# LINE 917 "Curses.hsc" #-}
    (406)         -> KeySUndo
{-# LINE 918 "Curses.hsc" #-}
    (407)       -> KeySuspend
{-# LINE 919 "Curses.hsc" #-}
    (408)          -> KeyUndo
{-# LINE 920 "Curses.hsc" #-}
    (410)        -> KeyResize
{-# LINE 921 "Curses.hsc" #-}
    _                          -> KeyUnknown (fromIntegral key)




getChHack :: IO Key
getChHack = liftM (decodeKey . fromIntegral . ord) getChar

getCh :: IO Key
getCh = threadWaitRead 0 >> (liftM decodeKey $ throwIfErr "getch" getch)

--getCh :: IO Key
--getCh = liftM decodeKey $ throwIfErr "getch" getch

-- getCh :: IO Key
-- getCh = do
--     nodelay stdScr 1
--     --halfdelay 1
--     v <- getch
--     case v of
-- 	(#const ERR) -> yield >> getCh 
-- 	x -> return $ decodeKey x
