{-# OPTIONS -#include "RSA_hsc.h" #-}
{-# LINE 1 "RSA.hsc" #-}
module RSA where
{-# LINE 2 "RSA.hsc" #-}



{-# LINE 5 "RSA.hsc" #-}

{-# LINE 6 "RSA.hsc" #-}

{-# LINE 7 "RSA.hsc" #-}

{-# LINE 8 "RSA.hsc" #-}

import Foreign
import CForeign
import MarshalArray
--import GHC.Weak
import Control.Exception as E
import IOExts

data RSAStruct 
data EVPCipherCtx
data EVPCipher
data EVP_PKEY
data EVP_CIPHER
data EVP_CIPHER_CTX
data EVP_MD_CTX
data EVP_MD
data BIGNUM

type RSA = Ptr RSAStruct
type EVPCIPHERCTX = Ptr EVPCipherCtx
type EVPCIPHER = Ptr EVPCipher

type EvpPkey = Ptr EVP_PKEY

------------------------
-- marshalling utilities
------------------------

throwZero_ s = throwIf_ (== 0) (const s)

withData :: [Word8] -> (Ptr CUChar -> CInt -> IO a) -> IO a
withData xs f = withArray xs (\a -> f (castPtr a) (fromIntegral $ length xs))

returnData :: Int -> (Ptr CUChar -> Ptr CInt -> IO ()) -> IO [Word8]
returnData sz f = do
	alloca (\bp -> allocaArray sz (\m -> f m bp >> peek bp >>= \s -> fmap (map fromIntegral) (peekArray (fromIntegral s) m)))

---------------
-- RSA routines
---------------


foreign import ccall unsafe "RSA_new" rsaNew :: IO RSA
foreign import ccall unsafe "RSA_free" rsaFree :: RSA -> IO ()

foreign import ccall unsafe "RSA_check_key" rsa_check_key :: RSA -> IO CInt

rsaCheckKey :: RSA -> IO ()
rsaCheckKey rsa = throwZero_ "RSA_check_key" $ rsa_check_key rsa


rsaSetN :: RSA -> Ptr BIGNUM -> IO ()
rsaSetN = ((\hsc_ptr -> pokeByteOff hsc_ptr 12))
{-# LINE 61 "RSA.hsc" #-}

rsaSetE :: RSA -> Ptr BIGNUM -> IO ()
rsaSetE = ((\hsc_ptr -> pokeByteOff hsc_ptr 16))
{-# LINE 64 "RSA.hsc" #-}

rsaSetD :: RSA -> Ptr BIGNUM -> IO ()
rsaSetD = ((\hsc_ptr -> pokeByteOff hsc_ptr 20))
{-# LINE 67 "RSA.hsc" #-}

rsaSetP :: RSA -> Ptr BIGNUM -> IO ()
rsaSetP = ((\hsc_ptr -> pokeByteOff hsc_ptr 24))
{-# LINE 70 "RSA.hsc" #-}

rsaSetQ :: RSA -> Ptr BIGNUM -> IO ()
rsaSetQ = ((\hsc_ptr -> pokeByteOff hsc_ptr 28))
{-# LINE 73 "RSA.hsc" #-}

rsaSetDMP1 :: RSA -> Ptr BIGNUM -> IO ()
rsaSetDMP1 = ((\hsc_ptr -> pokeByteOff hsc_ptr 32))
{-# LINE 76 "RSA.hsc" #-}

rsaSetDMQ1 :: RSA -> Ptr BIGNUM -> IO ()
rsaSetDMQ1 = ((\hsc_ptr -> pokeByteOff hsc_ptr 36))
{-# LINE 79 "RSA.hsc" #-}

rsaSetIQMP :: RSA -> Ptr BIGNUM -> IO ()
rsaSetIQMP = ((\hsc_ptr -> pokeByteOff hsc_ptr 40))
{-# LINE 82 "RSA.hsc" #-}




----------------------
-- Envelope Encryption 
----------------------

foreign import ccall unsafe "openssl/evp.h EVP_DigestInit" evpSignInit :: Ptr EVP_MD_CTX -> Ptr EVP_MD -> IO ()
foreign import ccall unsafe "openssl/evp.h EVP_DigestUpdate" evpSignUpdate :: Ptr EVP_MD_CTX -> Ptr CUChar -> CInt -> IO () 
foreign import ccall unsafe "openssl/evp.h EVP_SignFinal" evpSignFinal :: Ptr EVP_MD_CTX -> Ptr CUChar -> Ptr CInt -> Ptr EVP_PKEY -> IO Int 

foreign import ccall unsafe "openssl/evp.h EVP_OpenInit" evpOpenInit :: Ptr EVP_CIPHER_CTX -> Ptr EVP_CIPHER -> Ptr CUChar -> CInt -> Ptr CUChar -> Ptr EVP_PKEY -> IO CInt
foreign import ccall unsafe "openssl/evp.h EVP_DecryptUpdate" evpOpenUpdate :: Ptr EVP_CIPHER_CTX -> Ptr CUChar -> Ptr CInt -> Ptr CUChar -> CInt -> IO CInt 
foreign import ccall unsafe "openssl/evp.h EVP_OpenFinal" evpOpenFinal :: Ptr EVP_CIPHER_CTX -> Ptr CUChar -> Ptr CInt -> IO Int 

foreign import ccall unsafe "openssl/evp.h EVP_des_ede3_cbc" evpDesEde3Cbc :: IO (Ptr EVP_CIPHER)
foreign import ccall unsafe "openssl/evp.h EVP_md5" evpMD5 :: IO (Ptr EVP_MD)


foreign import ccall unsafe evpCipherContextBlockSize :: Ptr EVP_CIPHER_CTX -> IO Int 

{-# LINE 104 "RSA.hsc" #-}


evp_OpenInit :: Ptr EVP_CIPHER_CTX -> Ptr EVP_CIPHER -> [Word8] -> [Word8] -> Ptr EVP_PKEY -> IO ()
evp_OpenInit cctx cipher ek iv pkey =  do
	throwZero_ "EVP_OpenInit" $ withData ek (\a b -> withArray iv (\iv -> evpOpenInit cctx cipher a b (castPtr iv) pkey))
	return ()
	
evp_OpenUpdate :: Ptr EVP_CIPHER_CTX -> [Word8] -> IO [Word8]
evp_OpenUpdate cctx ind = do
    bsz <- evpCipherContextBlockSize cctx
    let sz = length ind + bsz
    withData ind (\ina inl -> returnData sz (\outa outl -> throwZero_ "EVP_OpenUpdate" (evpOpenUpdate cctx outa outl ina inl)))

evp_OpenFinal ::  Ptr EVP_CIPHER_CTX -> IO [Word8]
evp_OpenFinal cctx = do
    bsz <- evpCipherContextBlockSize cctx
    d <- returnData bsz (\outa outl -> throwZero_ "EVP_OpenFinal" (evpOpenFinal cctx outa outl))
    return d
    
    

foreign import ccall unsafe "EVP_PKEY_new" evpPKEYNew :: IO (Ptr EVP_PKEY)
foreign import ccall unsafe "EVP_PKEY_assign" evpPKEYAssign :: Ptr EVP_PKEY -> CInt -> Ptr a -> IO CInt 
foreign import ccall unsafe "EVP_PKEY_free" evpPKEYFree :: Ptr EVP_PKEY -> IO ()
foreign import ccall unsafe "EVP_PKEY_size" evpPKEYSize :: Ptr EVP_PKEY -> IO CInt

foreign import ccall unsafe "EVP_CIPHER_CTX_init" evpCipherCtxInit :: Ptr EVP_CIPHER_CTX -> IO CInt
foreign import ccall unsafe "EVP_CIPHER_CTX_cleanup" evpCipherCtxCleanup :: Ptr EVP_CIPHER_CTX -> IO CInt

withCipherCtx :: (Ptr EVP_CIPHER_CTX -> IO a) -> IO a
withCipherCtx action = allocaBytes (4216) $ \cctx -> 
{-# LINE 135 "RSA.hsc" #-}
	    E.bracket_ (throwZero_ "EVP_CIPHER_CTX_init" $ evpCipherCtxInit cctx) 
		(throwZero_ "EVP_CIPHER_CTX_cleanup" $ evpCipherCtxCleanup cctx) 
		    (action cctx)

withMdCtx :: (Ptr EVP_MD_CTX -> IO a) -> IO a
withMdCtx = allocaBytes (152) 
{-# LINE 141 "RSA.hsc" #-}
    
foreign import ccall unsafe pkeyNewRSA :: RSA -> IO (Ptr EVP_PKEY)

{-# LINE 144 "RSA.hsc" #-}

decryptAll :: [Word8] -> [Word8] -> Ptr EVP_PKEY -> [Word8] -> IO [Word8]
decryptAll keydata iv pkey xs = withCipherCtx $ \cctx -> do
	    d <- withData keydata $ \a b -> withArray iv $ \iv -> do
		throwZero_ "EVP_OpenInit" $ evpOpenInit cctx cipher a b (castPtr iv) pkey
		d <- evp_OpenUpdate cctx xs
		dr <- evp_OpenFinal cctx
		return $ d ++ dr  
	    return d where
	cipher = unsafePerformIO evpDesEde3Cbc

signAll :: Ptr EVP_PKEY -> [Word8] -> IO [Word8]
signAll  pkey xs = withMdCtx $ \cctx -> do
    evpSignInit cctx md5
    withData xs $ \a b -> (evpSignUpdate cctx a b)
    bsz <- fmap fromIntegral $ evpPKEYSize pkey
    d <- returnData bsz (\outa outl -> throwZero_ "EVP_SignFinal" (evpSignFinal cctx outa outl pkey))
    return d
	where md5 = unsafePerformIO evpMD5



--------------------
-- BigNum routines
--------------------

bn_bin2bn :: [Word8] -> IO (Ptr BIGNUM)
bn_bin2bn xs = throwIfNull "BN_bin2bn" $ withData xs (\a b -> bnBin2Bn a b nullPtr)

foreign import ccall unsafe "BN_bin2bn" bnBin2Bn :: Ptr CUChar -> CInt -> Ptr BIGNUM -> IO (Ptr BIGNUM)





readSymbolicLink :: FilePath -> IO FilePath
readSymbolicLink file =
  allocaBytes (4096) $ \buf -> do
{-# LINE 182 "RSA.hsc" #-}
    
    len <- withCString file $ \s ->
      throwErrnoIfMinus1 "readSymbolicLink" $
	c_readlink s buf (4096)
{-# LINE 186 "RSA.hsc" #-}
    peekCStringLen (buf,fromIntegral len)

foreign import ccall unsafe "unistd.h readlink"
  c_readlink :: CString -> CString -> CInt -> IO CInt

